'''Trains a simple convnet on the MNIST dataset.

Gets to 99.25% test accuracy after 12 epochs
(there is still a lot of margin for parameter tuning).
16 seconds per epoch on a GRID K520 GPU.

https://github.com/keras-team/keras/blob/master/examples/mnist_cnn.py

A Convolutional Network implementation example using TensorFlow library.
This example is using the MNIST database of handwritten digits
(http://yann.lecun.com/exdb/mnist/)

Author: Keras
Project: https://github.com/keras-team/keras

Modifications:
Modified to work with TF.Keras and to use IBM's Distributed Deep Learning library

*****************************************************************

Licensed Materials - Property of IBM

(C) Copyright IBM Corp. 2018. All Rights Reserved.

US Government Users Restricted Rights - Use, duplication or
disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

*****************************************************************
'''

from __future__ import print_function
from tensorflow.python import keras as keras
from tensorflow.python.keras.models import Sequential
from tensorflow.python.keras.layers import Dense, Dropout, Flatten
from tensorflow.python.keras.layers import Conv2D, MaxPooling2D
from tensorflow.python.keras.datasets import mnist
from tensorflow.python.keras import backend as K

import sys
import tensorflow as tf
import os
if len(sys.argv) > 1 and sys.argv[1] == '--eager':
    tf.compat.v1.enable_eager_execution()
import ddl

# Precreate parent directory needed for mnist data, to avoid collison between multiple processes
def setup_mnist():
    try:
        datadir = os.path.join(os.path.expanduser('~'), '.keras/datasets')
        if ddl.local_rank() == 0:
            if not os.path.exists(datadir):
                os.makedirs(datadir)
    except Exception as error:
        pass

    ddl.barrier()

batch_size = 128
num_classes = 10

epochs = 12

# input image dimensions
img_rows, img_cols = 28, 28

input_shape = ()
if K.image_data_format() == 'channels_first':
    input_shape = (1, img_rows, img_cols)
else:
    input_shape = (img_rows, img_cols, 1)

# DDL: Setup steps to avoid possible collison between processes.
setup_mnist()

# the data, split between train and test sets
(x_train, y_train), (x_test, y_test) = mnist.load_data()

if K.image_data_format() == 'channels_first':
    x_train = x_train.reshape(x_train.shape[0], 1, img_rows, img_cols)
    x_test = x_test.reshape(x_test.shape[0], 1, img_rows, img_cols)
    input_shape = (1, img_rows, img_cols)
else:
    x_train = x_train.reshape(x_train.shape[0], img_rows, img_cols, 1)
    x_test = x_test.reshape(x_test.shape[0], img_rows, img_cols, 1)
    input_shape = (img_rows, img_cols, 1)

x_train = x_train.astype('float32')
x_test = x_test.astype('float32')
x_train /= 255
x_test /= 255
print('x_train shape:', x_train.shape)
print(x_train.shape[0], 'train samples')
print(x_test.shape[0], 'test samples')

# convert class vectors to binary class matrices
y_train = keras.utils.to_categorical(y_train, num_classes)
y_test = keras.utils.to_categorical(y_test, num_classes)

model = Sequential()
model.add(Conv2D(32, kernel_size=(3, 3),
                 activation='relu',
                 input_shape=input_shape))
model.add(Conv2D(64, (3, 3), activation='relu'))
model.add(MaxPooling2D(pool_size=(2, 2)))
model.add(Dropout(0.25))
model.add(Flatten())
model.add(Dense(128, activation='relu'))
model.add(Dropout(0.5))
model.add(Dense(num_classes, activation='softmax'))


# DDL: adjust learning rate based on number of GPUs.
model.compile(loss=keras.losses.categorical_crossentropy,
              optimizer=tf.compat.v1.train.AdadeltaOptimizer(learning_rate=1.0 * ddl.size()),
              metrics=['accuracy'])

callbacks = list()

# DDL: Add the DDL callback.
callbacks.append(ddl.DDLCallback())
callbacks.append(ddl.DDLGlobalVariablesCallback())

# DDL: Only use verbose = 1 on rank 0.
model.fit(x_train, y_train,
          epochs=epochs,
          verbose=1 if ddl.rank() == 0 else 0,
          validation_data=(x_test, y_test),
          callbacks=callbacks)

# DDL: Only do final accuracy check on rank 0.
if ddl.rank() == 0:
    score = model.evaluate(x_test, y_test, verbose=0)
    print('Test loss:', score[0])
    print('Test accuracy:', score[1])
